%
%   PCC Standard Library
%
%   This file defines commands and keys used internally by PCC,
%   and a number of user-callable routines described in the
%   manual. Do not modify unless you are absolutely 100% positively
%   sure what you are doing.
%
%   This file will be overwritten by PCC upgrades. It will most
%   likely only work with the PCC version it was intended for.
%

%%% User-callable Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Suspend the process until the next turn. For restrictions of its
% use, see `Stop'.
Sub WaitOneTurn ()
  Local t = Turn
  Do While t = Turn
    Stop
  Loop
EndSub

% Usable only in ship context: set ship waypoint, and suspend until
% the ship arrives there. Special case for gravity wells. We terminate
% when the waypoint is zero: when moving across a map border, we never
% reach the specified x,y.
Sub MoveTo (x, y)
  Local pid = PlanetAt(x, y, 1)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    Do Until (Waypoint.DX=0 And Waypoint.DY=0) Or ((pid<>0) And (Orbit$=pid))
      Stop
    Loop
  EndWith
EndSub

Sub MoveTowards (x, y)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    WaitOneTurn
  EndWith
EndSub

% Enqueue a ship of type h (1..105), engine e (1..9) with
% bc beams of type bt, and tc torpedo launchers of type tt.
% Suspend until this order was enqueued successfully.
Sub EnqueueShip (h, e, Optional bt, bc, tt, tc)
  % wait until pending build order performed
  Do While Build.YesNo
    Stop
  Loop
  % wait until build order successfully delivered
  Do
    Try
      BuildShip h, e, bt, bc, tt, tc
      Break
    EndTry
    Stop
  Loop
EndSub

% Load specified cargo amount. Like `CargoUpload', but waits until
% cargo successfully loaded. Note that this will block forever if you
% specify more than you have, or are not at a planet.
Sub CargoUploadWait (cargo, Optional flags)
  Local Cargo.Remainder = cargo
  Do
    CargoUpload Cargo.Remainder, flags & "n"
    If Not Cargo.Remainder Then Return
    Stop
  Loop
EndSub

% Transfer specified cargo amount.
Sub CargoTransferWait (sid, cargo, Optional flags)
  Local Cargo.Remainder = cargo
  Do
    CargoTransfer sid, Cargo.Remainder, flags & "n"
    If Not Cargo.Remainder Then Return
    Stop
  Loop
EndSub

% Building with waiting
Sub BuildMinesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildMines Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildFactoriesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildFactories Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildBaseDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildBaseDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% Change this fleet's name
Sub SetFleetName (name)
  If Not IsEmpty(name) Then
    If Fleet$=0 Then Abort "Not in a fleet"
    Ships(Fleet$).Fleet.Name := name
  EndIf
EndSub

% Change leader of fleet fid to sid
Sub ChangeFleetLeader (fid, sid)
  If fid=sid Then Return
  % start the new fleet
  With Ships(sid) Do
    SetFleet sid       
    Fleet.Name := Ships(fid).Fleet.Name
  EndWith
  % First move all members to the new fleet, then move the old leader
  ForEach Ship Do If (Id<>fid) And (Fleet$=fid) Then SetFleet sid
  With Ships(fid) Do SetFleet sid
EndSub

% Switch to specified selection layer
Sub SelectionLayer (l)
  If IsEmpty(l) Then Return
  If IsString(l) Then
    Local c := Asc(l)
    If c>=97 And c<105 Then
      l := c-97
    Else
      If c>=65 And c<73 Then
        l := c-65
      Else
        Abort "Invalid selection layer"
      EndIf
    EndIf
  EndIf
  Selection.Layer := l
EndSub

% Fix ship.
Sub FixShipWait (Optional id)
  If IsEmpty(id) Then
    % assume ship context
    If (Not Orbit) Or (Not Planets(Orbit).Base.YesNo) Then Abort "Not at a starbase"
    Do While Planets(Orbit).Shipyard
      Stop
    Loop
    FixShip
  Else
    % assume starbase context
    Do While Shipyard
      Stop
    Loop
    FixShip id
  EndIf
EndSub

% Recycle ship.
Sub RecycleShipWait (Optional id)
  If IsEmpty(id) Then
    % assume ship context
    If (Not Orbit) Or (Not Planets(Orbit).Base.YesNo) Then Abort "Not at a starbase"
    Do While Planets(Orbit).Shipyard
      Stop
    Loop
    RecycleShip
  Else
    % assume starbase context
    Do While Shipyard
      Stop
    Loop
    RecycleShip id
  EndIf
EndSub

% Player notify from auto task. This function is magic.
Sub Notify (s, Optional t)
  If Not IsEmpty(t) Then
    Do While Turn < t
      Stop
    Loop
  EndIf
  t=1
  Do
    CC$Notify s
    Stop
  Loop While t
EndSub

% Deeper magic.
Sub CC$ScriptError (s)
  Notify "Error: " & s
  CC$Again
EndSub

% This function is still undocumented, because its interface is expected
% to change.
Sub AutoLoad (file, name, Optional args)
  % Strip "optional" from prototype
  Local call = args
  Local opt = Instr(args, "Optional ")
  If (opt=1) Or (opt>1 And (Mid(args, opt-1, 1)<=' ' Or Mid(args, opt-1, 1)=',')) Then
    call := Left(call, opt-1) & Mid(call, opt+9)
  EndIf
  Eval "Sub " & name & ("(" # args # ")"), "  Load '" & file & "'", "  " & name & (" " # call), "EndSub"
EndSub

%%% UI Functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

If System.GUI Then
%% When the GUI is not active, don't define all these to save memory
%% in text-mode PCC (aka ccs).

Dim Shared CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC

% Ctrl-P on starbase/ship build screen
Sub CC$PickupBuild
  If Build.YesNo Then
    CC$BuildH  := Build.Hull$
    CC$BuildE  := Build.Engine$
    CC$BuildBT := Build.Beam$
    CC$BuildBC := Build.Beam.Count
    CC$BuildTT := Build.Torp$
    CC$BuildTC := Build.Torp.Count
    Messagebox "Ship specification for a " & Build.Short & " saved. Use [Ctrl-B] on other starbases to build this ship there."
  EndIf
EndSub

% Ctrl-B on starbase screen
Sub CC$PerformBuild
  If IsEmpty(CC$BuildH) Or Owner$<>My.Race$ Then Return
  If Build.YesNo Then
    UI.Message "This starbase is already building a " & Build.Short & ". Override this with the saved build order for a " & Hulls(CC$BuildH).Name.Short & "?", "Ship Building", "Yes No"
    If UI.Result<>1 Then Return
  EndIf
  Try
    BuildShip CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC
  Else
    Messagebox "Unable to build a " & Hulls(CC$BuildH).Name.Short & ": " & System.Err
  EndTry
EndSub

% Ctrl-S on ship build screen (not yet documented because it's weird):
% search for a starbase which can perform this build order. This routine
% demonstrates all the inelegance of our language :-/
Sub CC$SearchBuild
  % Value of tech level t = 50*(t-1)*t
  If Build.YesNo Then
    CC$PickupBuild
    Local mstr := Hulls(Build.Hull$).Cost.Str
    Local tstr := "Max(0,Tech.Hull*(Tech.Hull-1)-" & (Hull(Build.Hull$).Tech.Hull*(Hull(Build.Hull$).Tech.Hull-1)) & ")"
    mstr := CAdd(mstr, CMul(Engine(Build.Engine$).Cost.Str, Hull(Build.Hull$).Engine.Count))
    tstr := tstr & "+Max(0,Tech.Engine*(Tech.Engine-1)-" & (Engine(Build.Engine$).Tech*(Engine(Build.Engine$).Tech-1)) & ")"
    If Build.Beam.Count Then
      mstr := CAdd(mstr, CMul(Beam(Build.Beam$).Cost.Str, Build.Beam.Count))
      tstr := tstr & "+Max(0,Tech.Beam*(Tech.Beam-1)-" & (Beam(Build.Beam$).Tech*(Beam(Build.Beam$).Tech-1)) & ")"
    EndIf
    If Build.Torp.Count Then
      mstr := CAdd(mstr, CMul(Torpedo(Build.Torp$).Cost.Str, Build.Torp.Count))
      tstr := tstr & "+Max(0,Tech.Torpedo*(Tech.Torpedo-1)-" & (Torpedo(Build.Torp$).Tech*(Torpedo(Build.Torp$).Tech-1)) & ")"
    EndIf
    UI.Search "Not Build.YesNo And CCompare(Mined.Str&(Supplies+Money)&'$','" & mstr & "'&(" & tstr & ")*50&'$')", "2b"
  EndIf
EndSub

Sub CC$ToggleSelection
  Mark Not Marked
EndSub

% F9 at various places
Sub CC$SetComment (obj)
  UI.Input "New comment for this " & obj & ":", "Comment for " & obj, 127, 380, Comment
  SetComment UI.Result
EndSub

% Ctrl-F9/Shift-F9
Sub CC$SetCommentAt (x, y, fl)
  Local pid = PlanetAt (x, y, fl)
  If pid Then With Planets(pid) Do CC$SetComment "Planet"
EndSub

% N in ship screen
Sub CC$ShipName
  If Owner$<>My.Race$ Then Return
  UI.Input "Enter new name:", "Rename Starship #" & Id, 20, 320, Name
  SetName UI.Result
EndSub

% N in fleet screen
Sub CC$FleetName
  UI.Input "Enter new fleet name:", "Rename Fleet #" & Fleet$, 250, 320, Ships(Fleet$).Fleet.Name
  SetFleetName UI.Result
EndSub

% B in fleet screen
Sub CC$ChangeLeader
  If Fleet$=Id Then Return      % ship is already the leader
  UI.Message "Make this ship the leader of the fleet?", "Fleet", "Yes No"
  If UI.Result=1 Then ChangeFleetLeader Fleet$, Id
EndSub

Sub CC$SetWaypoint
  SetWaypoint UI.X, UI.Y
EndSub

EndIf  % System.GUI

% HACK for revive.q compatibility. Since 1.1.13, the NewTurn hook is
% run by the PCC core. To avoid that the hook runs twice when user
% uses an old revive.q, abort it here. Because core.q is loaded first,
% nobody can run before us. Remove this in version 1.2.
Dim Shared CC$DidNewTurn
On Load Do CC$DidNewTurn = 0
On NewTurn Do If CC$DidNewTurn Then Return
On NewTurn Do CC$DidNewTurn = 1

%%% Initialisation of the Runtime System %%%%%%%%%%%%%%%%%%%%%%%%%%%%

% PCC assumes these statements are here
CreateShipProperty   Comment
CreatePlanetProperty Comment
CreateKeymap         Global
CreateKeymap         ControlScreen(Global)
CreateKeymap         ShipScreen(ControlScreen), PlanetScreen(ControlScreen)
CreateKeymap         BaseScreen(ControlScreen), FleetScreen(ControlScreen)
CreateKeymap         HistoryScreen(ControlScreen), AutoTaskScreen(ControlScreen)
CreateKeymap         ShipBuildScreen(Global), Starchart(Global)
Dim Shared           UI.Result
Dim Shared           UI.Directory
Dim Shared           UI.Prefix
Dim Shared           Cargo.Remainder
Dim Shared           Build.Remainder

% Commands handled internally by PCC.
% You are not expected to understand this 8-]
Bind ControlScreen   "f1"  := 101, "c-f1"  := 131, "s-f1"  := 141
Bind ControlScreen   "f2"  := 102, "c-f2"  := 132, "s-f2"  := 142
Bind ControlScreen   "f3"  := 103, "c-f3"  := 133, "s-f3"  := 143
Bind ControlScreen   "f4"  := 104, "c-f4"  := 134, "s-f4"  := 144
Bind ControlScreen   "f5"  := 105, "c-f5"  := 135, "s-f5"  := 145
Bind ControlScreen   "f6"  := 106, "c-f6"  := 136, "s-f6"  := 146
Bind ControlScreen   "f7"  := 107, "c-f7"  := 137, "s-f7"  := 147
Bind ControlScreen   "f8"  := 108, "c-f8"  := 138, "s-f8"  := 148
Bind ControlScreen   "f10" := 110, "c-f10" := 140, "s-f10" := 150
Bind ControlScreen   "pgup" := -6, "c-pgup" := 126, "wheelup" := -6
Bind ControlScreen   "pgdn" := -5, "c-pgdn" := 127, "wheeldn" := -5
Bind ControlScreen   "c-num-" := 126, "c-num+" := 127
Bind ControlScreen   "alt-c" := 161, "alt-q" := 162, "alt-r" := 160
Bind ControlScreen   "alt-k" := 166, "alt-n" := 167

Bind ShipScreen      "c-w" := 2, "tab" := 3, "s-tab" := 4, "c-tab" := 5, "d" := 23
Bind ShipScreen      "p" := 142
Bind FleetScreen     "c-w" := 2, "c-del" := 14, "ret" := 101
Bind BaseScreen      "u" := 10, "tab" := 1, "c-r" := 11, "c-c" := 13
Bind PlanetScreen    "u" := 10, "tab" := 1, "c-s" := 11

% Keys described in the help system and implemented here
Bind ControlScreen   "c-f9" := "CC$SetCommentAt UI.X, UI.Y, 1"
Bind ControlScreen   "s-f9" := "CC$SetCommentAt Loc.X, Loc.Y, 0"
Bind ControlScreen   "c-h" := "UI.Search Loc.X & ', ' & Loc.Y, 'spbuo4'"
Bind Starchart       "c-h" := "UI.Search UI.X & ', ' & UI.Y, 'spbuo4'"
Bind Starchart       "f7" := "UI.Search", "s" := "UI.Search"
Bind ShipScreen      "f9" := "CC$SetComment 'Ship'", "n" := "CC$ShipName"
Bind ShipScreen      "c-f" := "SetFCode Planet(Orbit$).FCode"
Bind ShipScreen      "c-z" := "SetWaypoint Loc.X, Loc.Y"
Bind PlanetScreen    "f9" := "CC$SetComment 'Planet'"
Bind BaseScreen      "f9" := "CC$SetComment 'Planet'"
Bind BaseScreen      "c-p" := "CC$PickupBuild", "c-b" := "CC$PerformBuild"
Bind FleetScreen     "f9" := "CC$SetComment 'Ship'", "n" := "CC$FleetName"
Bind FleetScreen     "b" := "CC$ChangeLeader"
Bind HistoryScreen   "f9" := "CC$SetComment 'Ship'"
Bind ShipBuildScreen "c-p" := "CC$PickupBuild", "c-s" := "CC$SearchBuild"

If System.GUI Then Print "[", System.Program, " ", System.Version, "]"
